# encoding: utf-8

"""
opcvm_sdk.services

@author:        Eric Pieuchot
@copyright:     Copyright 2016 Rubicubix Ltd
@license:       Apache-2.0
@contact:       contact@rubicubix.co.uk
"""

from opcvm_sdk.config import Config
from opcvm_sdk.models import Fundshare, History

from requests.auth import HTTPBasicAuth
from requests import Session

from threading import Thread

from re import match
from datetime import datetime
from dateutil.parser import parse

from simplejson import loads

from sys import exc_info


class ApiException(Exception):
    """OPCVM api Exception object"""

    def __init__(self, status=None, reason=None, response=None):
        if response:
            self.status = response.status_code
            self.reason = response.reason
            self.body = response.content
            self.headers = response.headers
        else:
            self.status = status
            self.reason = reason
            self.body = None
            self.headers = None

    def __str__(self):
        """Custom error message"""
        error_message = "Status: ({0}) Reason: {1}".format(
            self.status,
            self.reason
        )
        if self.headers:
            error_message += "\n HTTP Headers: {0}".format(
                self.headers
            )
        if self.body:
            error_message += "\n HTTP Body: {0}".format(
                self.body
            )
        return error_message


class ApiMixin(object):
    """OPCVM360 api mixin object"""

    def api_deserialize(self, data, klass=None):
        """
        Deserialize response data container

        Parameters
        ----------
        @type  data:     dict, list
        @param data:     data container
        @type  klass:    str
        @param klass:    class literal
        @rtype:          str, list, dict, object
        @return:         deserialized data
        """
        # Convert str to class
        if type(klass) == str:

            # Convert list
            if klass.startswith('list['):
                sub_kls = match('list\[(.*)\]', klass).group(1)
                return [self.api_deserialize(sub_data, sub_kls)
                        for sub_data in data]

            # Convert str to class
            klass = eval(klass)

        if klass in [int, long, float, str, bool]:
            return self._api_deserialize_primitive(data, klass)
        elif klass == datetime:
            return self._api_deserialize_datetime(data)
        elif klass == object:
            return data
        else:
            return self._api_deserialize_model(data, klass)

    def _api_deserialize_primitive(self, data, klass):
        """
        Deserialize string to primitive type.

        Parameters
        ----------
        @type  data:  str
        @param data:  parsed json
        @type  klass: str
        @param klass: class literal
        @rtype:       int, long, float, str, bool
        @return:      deserialized data
        """
        try:
            _value = klass(data)
        except UnicodeEncodeError:
            _value = unicode(data)
        except TypeError:
            _value = None
        return _value

    def _api_deserialize_datetime(self, string):
        """
        Deserialize string to datetime.

        Parameters
        ----------
        @type  string: str
        @param string: parsed json
        @rtype:        datetime
        @return:       deserialized date
        """
        try:
            _date = parse(string)
        except (ValueError, TypeError):
            _msg = "Failed to parse `{0}`".format(string)
            raise ApiException(status=0, reason=_msg)
        return _date

    def _api_deserialize_model(self, data, klass):
        """
        Deserialize list or dict to model.

        Parameters
        ----------
        @type  data:  dict
        @param data:  parsed json
        @type  klass: str
        @param klass: class literal
        @rtype:       opcvm_sdk.models.object
        @return:      deserialized object
        """
        if isinstance(data, dict):
            _instance = klass()
            for _attr, _attr_type in _instance._api_types:
                try:
                    _value = data[_attr]
                    setattr(
                        _instance,
                        _attr,
                        self.api_deserialize(_value, _attr_type)
                    )
                    del _value
                except KeyError:
                    _msg = "Missing attribute {0}".format(_attr)
                    raise ApiException(status=0, reason=_msg)
        else:
            _msg = "Incorrect type `{0}`".format(type(data))
            raise ApiException(status=0, reason=_msg)
        return _instance


class ApiClient(Session, ApiMixin):
    """OPCVM360 client api Session object"""

    def __init__(self, config=Config()):
        Session.__init__(self)
        self.config = config
        self._auth()
        self._header()

    def _header(self):
        """Add HTTP request header"""
        _headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        self.headers = _headers

    def _auth(self):
        """Encrypt HTTP request with username/password"""
        _username = self.config.username
        _password = self.config.password
        self.auth = HTTPBasicAuth(_username, _password)

    def api_call(self, method, resource, params=None,
                 klass=None, callback=None):
        """
        Makes the HTTP request (synchronous) and return the deserialized data.
        To make an async request, define a function for callback.

        Parameters
        ----------
        @type  method:   str
        @param method:   http request type
        @type  resource: str
        @param resource: method endpoint path
        @type  params:   dic
        @param params:   url query parameters
        @type  klass:    str
        @param klass:    class literal
        @type  callback: func
        @param callback: asynchronous request function
        @rtype:          dict
        @return:         deserialized http response
        """
        if callback is None:
            return self._api_call(
                method,
                resource,
                params,
                klass,
                callback
            )
        else:
            _thread = Thread(
                target=self.__api_call,
                args=(
                    method,
                    resource,
                    params,
                    klass,
                    callback
                )
            )
            _thread.start()
            return _thread

    def _api_call(self, method, resource, params=None,
                  klass=None, callback=None):
        """
        Commit, return and deserialized Response Object

        Parameters
        ----------
        @type  method:   str
        @param method:   http request type
        @type  resource: str
        @param resource: method endpoint path
        @type  params:   dict
        @param params:   url query parameters
        @type  klass:    str
        @param klass:    class literal
        @type  callback: func
        @param callback: asynchronous request function
        @rtype:          dict
        @return:         deserialized http response
        """
        # Request url
        _url = self.config.host + resource
        # Perform request and return response
        if str(method).upper() == 'GET':
            _response = self._api_call_get(_url, params)
        else:
            _msg = "Only GET http method available"
            raise ApiException(status=0, reason=_msg)
        # Unpack data container
        try:
            _data = loads(_response.content)
            _data = _data['data']
        except (KeyError, ValueError, AttributeError):
            _msg = "Missing data container in http response"
            raise ApiException(status=0, reason=_msg)
        # Deserialize data container
        if bool(_data):
            _data = self.api_deserialize(_data, klass)
        else:
            _msg = "Empty data container in http response"
            raise ApiException(status=0, reason=_msg)
        # Monitor threads
        if callback:
            callback(_data)
        else:
            return _data

    def _api_call_get(self, url, params=None):
        """
        Commit GET request and return Response Object

        Parameters
        ----------
        @type  url:    str
        @param url:    http request url
        @type  params: dict
        @param params: url query parameters
        @rtype:        requests.Response
        @return:       http response
        """
        # Perform http request
        try:
            _response = self.get(
                url,
                params=params,
                timeout=self.config.timeout
            )
        except:
            _msg = str(exc_info()[0])
            raise ApiException(status=0, reason=_msg)
        if _response.status_code != 200:
            raise ApiException(response=_response)
        # Return HTTP response
        return _response
