# encoding: utf-8

"""
opcvm_sdk.services

@author:        Eric Pieuchot
@copyright:     Copyright 2016 Rubicubix Ltd
@license:       Apache-2.0
@contact:       contact@rubicubix.co.uk
"""

from opcvm_sdk.config import Config
from opcvm_sdk.decorator import to_dataframe
from opcvm_sdk.client import ApiClient, ApiException

from dateutil.parser import parse


class OpcvmService(object):
    """OPCVM360 service object"""

    def __init__(self, config=Config()):
        self._api_client = ApiClient(config=config)

    def __del__(self):
        """Exit api client session"""
        self._api_client.__exit__()

    def get_fundshare_by_id(self, fund_id):
        """
        Get fundshare by id

        Parameters
        ----------
        @type  fund_id: int
        @param fund_id: fundshare id
        @rtype:         opcvm_sdk.models.Fundshare
        @return:        Fundshare object
        """
        # Set query parameters
        _resource = "{0}/{1}".format(
            self._api_client.config.fundshare,
            fund_id
        )
        _params = {
            'format': 'json',
        }
        # Get http request
        _fundshares = self._api_client.api_call(
            'GET',
            _resource,
            params=_params,
            klass='list[Fundshare]'
        )
        return _fundshares[0]

    def get_fundshare_by_isin(self, fund_isin):
        """
        Get fundshare by ISIN

        Parameters
        ----------
        @type  fund_isin: str
        @param fund_isin: fundshare ISIN
        @rtype:           opcvm_sdk.models.Fundshare
        @return:          Fundshare object
        """
        # ISIN sanity check
        _isinlength = self._api_client.config.isinlength
        if len(fund_isin) != _isinlength:
            _msg = "Unexpected ISIN length: {0} chars vs {1} ".format(
                len(fund_isin),
                _isinlength
            )
            raise ApiException(status=0, reason=_msg)
        # Set query parameters
        _resource = self._api_client.config.fundshare
        _params = {
            'queryFields': 'isin',
            'format': 'json',
            'query': fund_isin
        }
        # Get http request
        _fundshares = self._api_client.api_call(
            'GET',
            _resource,
            params=_params,
            klass='list[Fundshare]'
        )
        return _fundshares[0]

    @to_dataframe
    def get_history_by_id(self, fund_id, from_date, to_date):
        """
        Get fundshare history by id

        Parameters
        ----------
        @type  fund_id:   int
        @param fund_id:   fundshare id
        @type  from_date: str
        @param from_date: start date
        @type  to_date:   str
        @param to_date:   end data
        @type  only_updated:   bool
        @param only_updated:   history retrieved for updated data only
        @rtype:           pandas.DataFrame
        @return:          pandas DataFrame
        """
        # Time range sanity check
        _from = parse(from_date)
        _to = parse(to_date)
        if _from > _to:
            _msg = "Unordered dates: {0} > {1} ".format(
                from_date,
                to_date
            )
            raise ApiException(status=0, reason=_msg)
        # Set query parameters
        _resource = self._api_client.config.history
        _format = self._api_client.config.format
        _params = {
            'format': 'json',
            'fundshares': fund_id,
            'limit': int(1 + (_to - _from).days),
            'from': _from.strftime(_format),
            'to': _to.strftime(_format),
        }
        # Get http request
        _history = self._api_client.api_call(
            'GET',
            _resource,
            params=_params,
            klass='list[History]'
        )
        return _history

    def get_history_by_isin(self, fund_isin, from_date, to_date):
        """
        Get fundshare history by isin

        Parameters
        ----------
        @type  fund_isin: isin
        @param fund_isin: fundshare isin
        @type  from_date: str
        @param from_date: start date
        @type  to_date:   str
        @param to_date:   end data
        @type  only_updated:   bool
        @param only_updated:   history retrieved for updated data only
        @rtype:           pandas.DataFrame
        @return:          pandas DataFrame
        """
        _fundshare = self.get_fundshare_by_isin(fund_isin)
        _id = _fundshare.idFundShare
        del _fundshare
        _history = self.get_history_by_id(_id, from_date, to_date)
        return _history
